// 
// 29-Dec-19 R407 + 38g (DAI) 7.9A Evap -12 Comp 88
// 31-Dec-19 R407 +15g (DAI) 9.4A Exap    Comp 94
// 
#include "Weather.h"
#include "HPController.h"

#include <Adafruit_ADS1015.h>

// Adafruit_ADS1115 ads;  /* Use this for the 16-bit version */
Adafruit_ADS1015 ads;     /* Use thi for the 12-bit version */

Ticker tickerOSWatch, tickerDefrost, tickerTick, tickerMain;
bool bTick;

const char* version = "V1.0";
// USE_LITTLEFS
FS* filesystem = &LittleFS;

void checkPump();
void ICACHE_RAM_ATTR osWatch(void);
void mainCallback();
void defrostTickCallback();
void getSensorsCallback();
void setValveOff();
void weatherCallback();
void turnCompressorOn();

Task tValveOff(5000, TASK_FOREVER, &setValveOff);					//5 seconds
Task tCompressorOn(5000, TASK_FOREVER, &turnCompressorOn);			//5 seconds
Task tReadSensors(15000, TASK_FOREVER, &getSensorsCallback);		//15 seconds
Task tGetWeather(3600000, TASK_FOREVER, &weatherCallback);			//hourly
//Task tDewpoint(60000, TASK_FOREVER, &dewpointCallback);			//60 seconds

Scheduler runner;
ThermostatUDP thermostatUDP;

void doDefrost() { Defrost.bDoDefrost = true; }
void doTick() { bTick = true; }
void doMain() { ss.bMain = true; }

//For MQTT
#define RECONNECT_DELAY_W 5
uint32_t elapsed = 0;
PangolinMQTT mqttClient;
Ticker mqttReconnectTimer;

//void printDirectory(File dir, int numTabs) {
//	while (true) {
//
//		File entry = dir.openNextFile();
//		if (!entry) {
//			// no more files
//			break;
//		}
//		for (uint8_t i = 0; i < numTabs; i++) {
//			Serial.print('\t');
//		}
//		Serial.print(entry.name());
//		if (entry.isDirectory()) {
//			Serial.println("/");
//			printDirectory(entry, numTabs + 1);
//		}
//		else {
//			// files have sizes, directories do not
//			Serial.print("\t\t");
//			Serial.println(entry.size(), DEC);
//		}
//		entry.close();
//	}
//}
//
//////////////////////////////////////////////////////////////////////////////////////////
//This function in main cpp or ino to set the system time
void time_is_set(void)
{
	//Serial.println("------------------ settimeofday() was called ------------------");
	ntp.setSystemTime();
	connectToMqtt();
	ss.timeSynchronized = true;
	ss.bGetWeather = true;
}

////////////////////////////////////////////////////////  MQTT stuff    /////////////////////////////////////////////////
boolean isValidNumber(String str) {
	boolean isNum = false;
	for (byte i = 0; i < str.length(); i++)
	{
		isNum = isDigit(str.charAt(i)) || str.charAt(i) == '+' || str.charAt(i) == '.' || str.charAt(i) == '-';
		if (!isNum) return false;
	}
	return isNum;
}

void connectToMqtt() {
	DEBUG_PRINTLN("Connecting to MQTT...");
	mqttClient.connect();
}

void onMqttConnect(bool sessionPresent) {
	elapsed = millis();
	ss.mqttActive = true;
	DEBUG_PRINTF("Connected to MQTT session=%d max payload size=%d\n", sessionPresent, mqttClient.getMaxPayloadSize());
	DEBUG_PRINTLN("Subscribing at QoS 2");
	mqttClient.subscribe(ss.topic, 2);

	String jsonString = ss.getSystemStatus(false);
	mqttClient.publish(ss.topicr, jsonString.c_str(), jsonString.length(), 0, false);
}

void onMqttDisconnect(int8_t reason) {
	Serial.printf("Disconnected from MQTT reason=%d\n", reason);
	ss.mqttActive = false;

	if (WiFi.isConnected()) {
		mqttReconnectTimer.once(2, connectToMqtt);
	}
}

void onMqttMessage(const char* topic, const uint8_t* payload, size_t len, uint8_t qos, bool retain, bool dup) {
	char message[100] = {};

	DEBUG_PRINTF("T=%u Message %s qos%d dup=%d retain=%d len=%d elapsed=%u\n", millis(), topic, qos, dup, retain, len, millis() - elapsed);
	//PANGO::dumphex(payload, len, 16);
	//DEBUG_PRINTF("JSON:%s\n", payload);
	const size_t capacity = JSON_OBJECT_SIZE(20) + 100;
	DynamicJsonDocument doc(capacity);
	deserializeJson(doc, payload);

	// Test if parsing succeeds.
	if (doc.isNull()) {
		Serial.println("parseObject() failed");
		return;
	}

	String cmd = doc["cmd"];

	DEBUG_PRINTF("MQTT:Cmd:%s\n", cmd.c_str());

	if (cmd == F("status")) {
		String jsonString = ss.getSystemStatus(false);
		mqttClient.publish(ss.topicr, jsonString.c_str(), jsonString.length(), 0, false);
		//mqttClient.publish(ss.topicr, 0, false, ss.getSystemStatus(false));
		return;
	}
	else if (cmd == F("setT")) {
		String tempSet = doc["temp"];
		String targetT = doc["targetT"];
		if (isValidNumber(tempSet)) {
			DEBUG_PRINTF("Set temp:%s\n", tempSet.c_str());
			ss.setRoomTemperature = tempSet.toInt();
			sprintf(message, G("Temperature set to :%d"), ss.setRoomTemperature);
			eepromUtil.eeprom_write_byte(E_ROOMTEMP, ss.setRoomTemperature);
		}
		if (isValidNumber(targetT)) {
			DEBUG_PRINTF("Set targetT:%s\n", targetT.c_str());
			ss.targetT = targetT.toInt();
			sprintf(message, G("targetT set to :%d"), ss.targetT);
		}

		if (ss.operatingMode == COOL) {
			eepromUtil.eeprom_write_int(E_COOLTEMP, ss.targetT);
			ss.coolT = ss.targetT;
		}
		else {
			eepromUtil.eeprom_write_int(E_HEATTEMP, ss.targetT);
			ss.heatT = ss.targetT;
		}

		EEPROM.commit();
	}

	//{"cmd":"save","onH":23,"offH":13,"mode":true}
	else if (cmd == F("save")) {
		ss.startTime = doc["startTime"]; // 23
		ss.stopTime = doc["stopTime"]; // 13
		ss.targetT = doc["targetT"];
		ss.bUseWeather = doc["useWeather"];
		ss.operatingMode = doc["operatingMode"]; // true
		eepromUtil.eeprom_write_byte(E_USEWEATHER, ss.bUseWeather);
		eepromUtil.eeprom_write_int(E_STARTTIME, ss.startTime);
		eepromUtil.eeprom_write_int(E_STOPTIME, ss.stopTime);
		if (ss.operatingMode == COOL) {
			eepromUtil.eeprom_write_int(E_COOLTEMP, ss.targetT);
			ss.coolT = ss.targetT;
		}
		else {
			eepromUtil.eeprom_write_int(E_HEATTEMP, ss.targetT);
			ss.heatT = ss.targetT;
		}
		
		EEPROM.commit();

		DEBUG_PRINTF(G("Start time set to: %d\nStop time set to: %d"), ss.startTime, ss.stopTime);
	}

	else if (cmd == F("on")) {
		bool newMode = ss.masterOnOff;
		ss.masterOnOff = ON;
		if (newMode != ss.masterOnOff)
		{
			ss.powerOnOff();
			ss.setAuto(false);
			sprintf(message, "%s", F("Heatpump is ON"));
		}
	}

	else if (cmd == F("off")) {
		bool newMode = ss.masterOnOff;
		ss.masterOnOff = OFF;
		if (newMode != ss.masterOnOff)
		{
			ss.powerOnOff();
			ss.setAuto(false);
			sprintf(message, "%s", F("Heatpump is OFF"));
		}
	}

	else if (cmd == F("auto")) {
		bool newMode = ss.masterOnOff;
		ss.masterOnOff = ON;
		if (newMode != ss.masterOnOff)
		{
			ss.powerOnOff();
			ss.setAuto(true);
			sprintf(message, "%s", F("Set to auto"));
		}
	}

	//else if (cmd == F("ip")) {
	//	const char* ip = doc["ip"];
	//	strncpy(ss.ipaddress, ip, 15);
	//	ss.ipaddress[16] = '\0';
	//	eepromUtil.eeprom_write_string(E_IPADD, ss.ipaddress);
	//	DEBUG_PRINTF("IP:%s\n", ss.ipaddress);
	//	EEPROM.commit();
	//}

	else if (cmd == F("manual")) {
		ss.setAuto(false);
		sprintf(message, "%s", F("Set to manual"));
	}

	else if (cmd == F("reboot"))
	{
		sprintf(message, "%s", F("Rebooting"));
		delay(1000);
		ss.shouldReboot = true;
	}

	//Return status message
	if (strlen(message) > 0)
	{
		const size_t capacity = JSON_OBJECT_SIZE(1) + 220;
		DynamicJsonDocument root(capacity);
		String jsonString;

		root["message"] = message;
		serializeJson(root, jsonString);

		mqttClient.publish(ss.topicr, jsonString.c_str(), jsonString.length(), 0, false);
	}

	ss.bMain = true;							//Force update
	DEBUG_PRINTF("MQTT End: %s\n", message);
}

////////////////////////////////////////////////////////   Setup  /////////////////////////////////////////////////

File root;
void setup() {
	char result[16];

	Serial.begin(115200);
	delay(2000);
	Serial.println("Startup");

	// Format FileFS if not yet
	if (!FileFS.begin())
	{
		Serial.print(FS_Name);
		Serial.println(F(" failed! AutoFormatting."));
		FileFS.format();
	}
	digitalWrite(PUMPGPIO, LOW);
	pinMode(PUMPGPIO, OUTPUT);
	digitalWrite(PUMPGPIO1, HIGH);
	pinMode(PUMPGPIO1, OUTPUT);

	//Test a GPIO output
	//do {
	//	digitalWrite(PUMPGPIO, LOW);
	//	delay(1000);
	//	digitalWrite(PUMPGPIO, HIGH);
	//	delay(1000);
	//} while (1);
	Wire.begin(4, 5);							//sda, scl
	Wire.setClockStretchLimit(15000);
	ss.expanderWrite(0xff);						//Set everything off just in case

	// Setup ntp and install callback - called when settimeofday is called (by SNTP or us) once enabled (by DHCP), SNTP is updated every hour
	ntp.setupTime();
	settimeofday_cb(time_is_set);

	DEBUG_PRINTF("%s Setup...\n", version);

	ss.readSettings();						//Read settings from EEPROM
	//eepromUtil.eeprom_write_byte(E_DHCP, 1);					//Set to DHCP
	//EEPROM.commit();

	DEBUG_PRINTLN(G("Startup wifi"));
	char hostName[16] = "Heatpump";
	if (strlen(ss.hostName) > 0)
		snprintf(hostName, sizeof(hostName), "Heatpump-%s", ss.hostName);

	WiFi.hostname(hostName);
	wifiWPS.setupWifi();
	DEBUG_PRINTF("WiFi connected!\n%d.%d.%d.%d\n", WiFi.localIP()[0], WiFi.localIP()[1], WiFi.localIP()[2], WiFi.localIP()[3]);
	debugWeb.wifiActive = true;

	setupWebServer();
	setupOTAUpdate();

	ss.millisNow = millis();							//Save current millis in case we need a startup delay

	if (ss.operatingMode == COOL)
		ss.targetT = ss.coolT;
	else
		ss.targetT = ss.heatT;
	sensors.readTemperatureSensors();
	thermostatUDP.initializeUDP();

	/* Start timers for system functions */
	runner.init();
	runner.addTask(tReadSensors);
	runner.addTask(tValveOff);
	runner.addTask(tGetWeather);
	runner.addTask(tCompressorOn);

	tReadSensors.enable();
	tGetWeather.enableDelayed();

	//Set up watchdog timer
	ss.last_loop = millis();
	tickerOSWatch.attach_ms(((OSWATCH_RESET_TIME / 3) * 1000), osWatch);
	tickerTick.attach(1, doTick);
	tickerMain.attach(10, doMain);

	//Setup and connect to MQTT
	mqttClient.onConnect(onMqttConnect);
	mqttClient.onDisconnect(onMqttDisconnect);
	mqttClient.onMessage(onMqttMessage);
	mqttClient.setCredentials(MQTT_USER, MQTT_PASS);
	mqttClient.setServer(MQTT_HOST, MQTT_PORT);
	mqttClient.setCleanSession(true);

	DEBUG_PRINTLN(G("Setup done"));

	ads.begin();

	////Initialize SD card
	//if (!SD.begin(CS_PIN)) {
	//	Serial.println("SD failed");
	//}
	//else
	//{
	//	Serial.println("SD initialized");
	//	root = SD.open("/");

	//	printDirectory(root, 0);
	//	Serial.println("File list done!");
	//}
}

void readads(void)
{
	int16_t adc0, adc1, adc2, adc3;

	adc0 = ads.readADC_SingleEnded(0);
	adc1 = ads.readADC_SingleEnded(1);
	adc2 = ads.readADC_SingleEnded(2);
	adc3 = ads.readADC_SingleEnded(3);
	Serial.print("AIN0: "); Serial.println(adc0);
	Serial.print("AIN1: "); Serial.println(adc1);
	Serial.print("AIN2: "); Serial.println(adc2);
	Serial.print("AIN3: "); Serial.println(adc3);
	Serial.println(" ");
}

////////////////////////////////////////////////////////  Loop    /////////////////////////////////////////////////
void loop() {

	//readads();
	ArduinoOTA.handle();
	runner.execute();
	ss.last_loop = millis();							//Timer for watchdog

	if (Defrost.bDoDefrost) {
		defrostTickCallback();
		Defrost.bDoDefrost = false;
	}

	if (bTick)
	{
		String timeNow = ntp.getLocalTime();
		ss.sendWsMessage(timeNow, "timestamp"); //send event "timestamp"
		thermostatUDP.getUDPJson();
		bTick = false;
	}

	//Delay valve off 
	if (ss.bValveOff)
	{
		tValveOff.enableDelayed();
		ss.bValveOff = false;
	}

	if (ss.bMain)
	{
		mainCallback();
		ss.bMain = false;
	}

	if (ss.bGetWeather) {
		if (WiFi.isConnected())
			Weather.getWeather();
		ss.bGetWeather = false;
	}

	//Save network settings
	if (ss.bSaveNetSettings) {
		ss.saveNetSettings();
		ss.bSaveNetSettings = false;
	}

	//Save  settings
	if (ss.bSaveSettings) {
		ss.saveSettings();
		ss.bSaveSettings = false;
	}

	if (ss.shouldReboot) {
		DEBUG_PRINTLN(G("Rebooting..."));
		delay(100);
		ESP.restart();
	}

	if (ss.refreshData)
	{
		//Send json with system status to web browser
		ss.sendSystemStatus();
		ss.refreshData = false;
	}

}

////////////////////////////////////////////////////////  Callback functions   /////////////////////////////////////////////////
void setValveOff() {
	ss.setRelay(VALVE, OFF);
	tValveOff.disable();
}

void getSensorsCallback() {
	sensors.readTemperatureSensors();
	thermostatUDP.sendStatus();
}

void defrostTickCallback() {
	Defrost.defrostTimer++;
	if (Defrost.defrostStep == 0)
		ss.sendWsMessage("Start:" + (String)(Defrost.defrostDelay - Defrost.defrostTimer), "deftime"); //send event "deftime"
	else
		ss.sendWsMessage("Remain:" + (String)(Defrost.defrostTimeToNextStep - Defrost.defrostTimer) + "(" + Defrost.defrostStep + ")", "deftime"); //send event "deftime"

	Defrost.defrostCheck();
}

void weatherCallback() {
	if (WiFi.isConnected())
		Weather.getWeather();
}

void turnOnHeat() {
	DEBUG_PRINTLN(G("Turning on - HEAT"));
	ss.setRelay(FAN, ON);
	ss.setRelay(VALVE, ON);
	tCompressorOn.enableDelayed();
}

void turnCompressorOn(){
	tCompressorOn.disable();
	ss.setRelay(COMPRESSOR, ON);
	ss.setOperatingState(ON);
	DEBUG_PRINTLN(G("Turning compressor on"));
}

void ICACHE_RAM_ATTR osWatch(void) {
	unsigned long t = millis();
	unsigned long last_run = abs(t - ss.last_loop);
	if (last_run >= (OSWATCH_RESET_TIME * 1000)) {
		// save the hit here to eeprom or to rtc memory if needed
		ESP.restart();  // normal reboot 
		//ESP.reset();  // hard reset
	}
}

//Check circulating pump and turn on/off as necessary
void checkPump() {
	bool set;

	if (ss.masterOnOff == ON) {
		if (ss.currentOperatingState == ON)
		{
			//DEBUG_PRINTLN(G("Turning pump on"));
			set = ON;
		}
		else
		{
			if (ss.bRunPumpWhenOff)
			{
				set = ON;
			}
			else
			{
				set = OFF;
			}
		}
	}
	else            //HP is off
	{
		set = OFF;
	}

	ss.setRelay(PUMP, set);
}


////////////////////////////////////////////////////////  Main logic    /////////////////////////////////////////////////
void mainCallback()
{
	ss.setRelay(0, true);					//Update relays every cycle in case the update didn't work for some reason
	checkPump();

	if (ss.startupDelay)
	{
		DEBUG_PRINTF("delay:%ld\n", (ss.millisNow + 120000) - (ss.millisNow + millis()));
		if (ss.millisNow + millis() > ss.millisNow + 120000)
			ss.startupDelay = false;
		return;
	}

	//If master on/off is off turn everything off
	if (ss.masterOnOff == OFF)
	{
		if (ss.currentOperatingState == ON)
			ss.turnOff();
	}
	else
	{
		if (ss.automatic) {
			if (!ss.getTimeBetween(ntp.hourT, ss.startTime, ss.stopTime)) {
				ss.turnOff();
				return;
			}
		}

		if (((ss.currentOperatingState == OFF) && (ss.masterOnOff == OFF)) || (Defrost.defrostFlag))
		{
			ss.refreshData = true;
			return;
		}

		//If compressor temperature exceeds maximum turn the compressor off
		if (TEMP_COMPRESSOR > ss.maxCompTemperature)
		{
			ss.setRelay(COMPRESSOR, OFF);
			ss.bValveOff = true;			//Delay for valve off
			ss.setRelay(FAN, OFF);
			ss.overheat = true;
			DEBUG_PRINTLN(G("Compressor overheat - turn off"));
		}

		if ((TEMP_COMPRESSOR < TEMP_OVERHEAT_RESET) && ss.overheat)
		{
			ss.overheat = false;				//Reset overheat condition
			ss.currentOperatingState = OFF;			//Set so the compressor re-starts
			DEBUG_PRINTLN(G("Compressor cooled"));
		}

		////////////////////////////////////////// For heat /////////////////////////////////////////////////////////////////////////////
		if (ss.operatingMode == HEAT) {
			if (ss.currentOperatingState == ON) {
				if ((TEMP_WATER >= ss.targetT) || ((ss.thermostatTemperature >= ss.setRoomTemperature) && ss.useThermostat)) {
					DEBUG_PRINTLN(G("Target reached turning off - HEAT"));
					//If we have achieved the target temperature then turn everything off
					ss.setRelay(COMPRESSOR, OFF);
					ss.bValveOff = true;			//Delay for valve off
					ss.setRelay(FAN, OFF);
					ss.setOperatingState(OFF);
					ss.sendWsMessage("Hysterisis delay", "hysterisis");
				}

				//Check to see if we need to do a defrost. 
				if (TEMP_EVAPORATOR <= Defrost.defrostTemperature) {
					//Check for last defrost time
					ulong mn = millis();
					ulong mtc = Defrost.defrostTimeBetween * 60 * 1000;
					DEBUG_PRINTF(G("Last:%d Now:%d Between:%d\n"), Defrost.lastDefrost, mn, mtc);

					if ((mn - Defrost.lastDefrost >= mtc) || (TEMP_EVAPMID < -8 && Defrost.lastDefrost < 50000))
					{
						if (!Defrost.defrostFlag)
						{
							DEBUG_PRINTLN(G("Start defrost cycle"));
							Defrost.defrostTimer = 0;
							Defrost.lastDefrostStep = 0;
							Defrost.defrostFlag = true;
							tickerDefrost.attach(1, doDefrost);
						}
					}
				}
			}
			else //Operating state is OFF
			{
				//Operational modes for heating. Override hysterisis if just turned on
				int hysterisis;
				if (ss.justStarted) {
					ss.justStarted = false;
					hysterisis = 0;
				}
				else
					hysterisis = ss.hysterisis;

				if (hysterisis != 0) {
					DEBUG_PRINTLN(G("Wait on hysterisis drop"));
				}

				//			if ((TEMP_WATER <= ss.targetT - hysterisis) || ((ss.thermostatTemperature <= ss.setRoomTemperature - hysterisis) && ss.useThermostat)) {
				if (TEMP_WATER <= ss.targetT - hysterisis) {
					if (ss.useThermostat) {
						if (ss.thermostatTemperature <= ss.setRoomTemperature - hysterisis)
						{
							turnOnHeat();
						}
					}
					else {	//Not using thermostat. Only turn on if water temp lower than required
						if (TEMP_WATER < ss.targetT){
							turnOnHeat();
						}
					}
				}
			}
		}

		//////////////////////////////////////// For cooling ///////////////////////////////////////////////////////////////////////////////
		if (ss.operatingMode == COOL) {
			if (ss.currentOperatingState == ON) {
				//Operational modes for cooling
				if (TEMP_WATER <= ss.coolT) {
					//If we have achieved the target temperature then turn everything off
					DEBUG_PRINTLN(G("Target reached turning off - COOL"));
					ss.setRelay(COMPRESSOR, OFF);
					ss.setRelay(VALVE, OFF);
					ss.setRelay(FAN, OFF);
					ss.setOperatingState(OFF);
				}
			}
			else
			{
				//Operational modes for cooling
				if (TEMP_WATER >= ss.coolT + ss.hysterisis) {
					DEBUG_PRINTLN(G("Turning on - COOL"));
					ss.setRelay(VALVE, OFF);
					ss.setRelay(FAN, ON);
					tCompressorOn.enableDelayed();
				}
			}
		}
	}
	ss.refreshData = true;
}

